from simple_salesforce import Salesforce

class SFDC_Helper:

	def __init__(self, p_creds, p_sw):
		self.sf                    = None
		self.sw                    = p_sw
		self.object_labels         = {}
		self.object_desc           = {}
		self.fx_rates              = {}
		self.table_data            = {}
		self.table_data_cols       = {}
		self.conected              = False
		self.standard_object_names = set()
		self.custom_object_names   = set()
		self.custom_setting_names  = set()
		self.userid                = p_creds ['user']   if 'user'   in p_creds.keys() else None
		self.password              = p_creds ['passwd'] if 'passwd' in p_creds.keys() else None
		self.token                 = p_creds ['token']  if 'token'  in p_creds.keys() else None
		
	def connect (self):
		self.sw.echo ("Connecting to SFDC with Userid [" + self.userid + "]")
		self.sf = Salesforce(username=self.userid, password=self.password, security_token=self.token)
		self.sw.echo ("Connected")
	
	def load_metadata (self):
		if len(self.object_labels.keys()) == 0:
			self.sw.echo ("Scanning Standard & Custom Objects")
			for x in self.sf.describe()["sobjects"]:
				api_name = x["name"]
				label = x["label"]
				isCustomSetting = x["customSetting"]
				isCustomObject = x["custom"]
				self.object_labels[api_name] = label
				if isCustomSetting:
					self.custom_setting_names.add (api_name)
				elif isCustomObject:
					self.custom_object_names.add (api_name)
				else:
					self.standard_object_names.add (api_name)

	def describe_object (self, api_name):
		if api_name not in self.object_desc.keys():
			self.sw.echo ("Loading SFDC Metadata for [" + api_name + "]")
			resp = getattr(self.sf, api_name).describe()
			self.object_desc[api_name] = resp['fields']
		return self.object_desc[api_name]

	def load_fx_rates (self):
		self.sw.echo ("Loading Currency Rates..")
		qs = "SELECT ISOCode, ConversionRate FROM CurrencyType WHERE IsActive=TRUE"
		resp = self.sf.query_all (qs)
		loopIndex = 0
		while loopIndex < len(resp['records']):
			self.fx_rates [resp['records'][loopIndex]['IsoCode']] = resp['records'][loopIndex]['ConversionRate']
			loopIndex += 1
		self.sw.echo (str(len(self.fx_rates.keys())) + " Currency Rates Loaded")

	def get_amount_in_USD (self, amount, curreny_code):
		if curreny_code in self.fx_rates.keys():
			return (amount / self.fx_rates[curreny_code])
		return 0.0

	def split_list_into_array_of_lists (self, inlist, arraysize=500):
		array_of_lists = []
		list_items = []
		for member in inlist:
			if len(list_items) >= arraysize:
				array_of_lists.append (list_items)
				list_items = list()
				list_items.append (member)
			else:
				list_items.append (member)
		if len(list_items) > 0:
			array_of_lists.append (list_items)
		return array_of_lists;

	def load_data (self, objectName, colnames, filters, joincondition=[], limit=0, orderby=''):
		self.sw.echo ("Selecting from " + objectName + "..")
		if len(colnames) == 0:
			if objectName not in self.object_desc.keys():
				for field in self.describe_object (objectName):
					colnames.append (field['name'])
		self.table_data[objectName] = []
		self.table_data_cols[objectName] = colnames
		query_list = []
		qs = 'SELECT '
		qs += ','.join (colnames)
		qs += ' FROM ' + objectName
		if (len(filters) > 0):
			qs += ' WHERE ' 
			qs += ' AND '.join (filters)
		if len(orderby) > 0:
			qs += ' ORDER BY ' + orderby
		if limit > 0:
			qs += ' LIMIT ' + str(limit)
		if len(joincondition) == 3:
			joinField = joincondition[0]
			joinedObject = joincondition[1]
			joinedObjectCol = joincondition[2]
			joinVals = []
			for rec in self.table_data[joinedObject]:
				joinVals.append ("'" + rec[joinedObjectCol] + "'")
				if len(joinVals) >= 500:
					qsj = qs + ' AND ' + joinField + ' IN (' + ",".join(joinVals) + ")"
					query_list.append (qsj)
					joinVals = []
			if len(joinVals) > 0:
				qsj = qs + ' AND ' + joinField + ' IN (' + ",".join(joinVals) + ")"
				query_list.append (qsj)
		else:
			query_list.append (qs)

		for qs in query_list:
			resp = self.sf.query_all(qs)
			arrLen = len(resp['records'])
			loopIndex = 0
			while loopIndex < arrLen:
				rec = {}
				for col in colnames:
					if '.' in col:
						colparts = col.split ('.')
						if len(colparts) == 2:
							rec[col] = resp['records'][loopIndex][colparts[0]][colparts[1]]
					else:
						rec[col] = resp['records'][loopIndex][col]
				loopIndex += 1
				self.table_data[objectName].append (rec)
			self.sw.echo (str(len(self.table_data[objectName])) + " rows Selected from " + objectName)

	def calculate_usd_column (self, objectName, amtcol, currencycol, newcol):
		if amtcol in self.table_data_cols[objectName] and currencycol in self.table_data_cols[objectName]:
			newrecs = []
			if newcol not in self.table_data_cols[objectName]:
				self.table_data_cols[objectName].append (newcol)
			for rec in self.table_data [objectName]:
				rec[newcol] = self.get_amount_in_USD (rec[amtcol], rec[currencycol])
				newrecs.append (rec)
			self.table_data[objectName] = newrecs



name = "salesforce_dbsync"

from salesforce_dbsync.class_SFDC_Helper import SFDC_Helper
from salesforce_dbsync.class_MySQL_Helper import MySQL_Helper
from salesforce_dbsync.class_Sync_Manager import Sync_Manager



import mysql.connector

class MySQL_Helper:

	def __init__ (self, p_mysql_creds, p_sw=None):
		self.host    = p_mysql_creds ['host']   if 'host'   in p_mysql_creds.keys() else 'localhost'
		self.user    = p_mysql_creds ['user']   if 'user'   in p_mysql_creds.keys() else None
		self.passwd  = p_mysql_creds ['passwd'] if 'passwd' in p_mysql_creds.keys() else None
		self.db      = p_mysql_creds ['db']     if 'db'     in p_mysql_creds.keys() else None
		self.sw      = p_sw
		self.connected = False
		self.connected_to_db = False
		self.tables  = []

	def connect (self):
		if self.host and self.user and self.passwd:
			self.sw.echo ("Connecting to MySQL. Host[" + self.host + "], User[" + self.user + "]")
			self.mysql = mysql.connector.connect (host=self.host, user=self.user, passwd=self.passwd)
			self.connected = True
			self.connected_to_db = False
			self.sw.echo ("MySQL Connected")
		else:
			self.sw.echo ("ERROR: Missing one or more of [host, user, passwd]")

	def connect_db (self):
		if self.host and self.user and self.passwd and self.db:
			self.sw.echo ("Connecting to MySQL Database. Host[" + self.host + "], User[" + self.user + "], DB[" + self.db + "]")
			self.mysql = mysql.connector.connect (host=self.host, user=self.user, passwd=self.passwd, database=self.db)
			self.connected = False
			self.connected_to_db = True
			mycursor = self.mysql.cursor (buffered=True)
			mycursor.execute ("SHOW TABLES")
			for x in mycursor:
				self.tables.append (x[0].lower())
			self.sw.echo ("MySQL Database Connected. Database has " + str(len(self.tables)) + " Tables")
		else:
			self.sw.echo ("ERROR: Missing one or more of [host, user, passwd, db]")

	def create_database_if_not_exists (self):
		if not self.connected:
			self.connect ()
		if self.connected:
			mycursor = self.mysql.cursor (buffered=True)
			mycursor.execute ("SHOW DATABASES")
			for x in mycursor:
				if (x[0] == self.db):
					self.sw.echo ("Found Pre-Existing Database [" + x[0] + "]")
					return True
			mycursor.execute ("CREATE DATABASE " + self.db)
			self.sw.echo ("Created Database [" + self.db + "]")

	def table_exists (self, table_name):
		for tabname in self.tables:
			if tabname.lower() == table_name.lower():
				self.sw.echo ("Found Pre-Existing Table [" + tabname + "]")
				return True
		self.sw.echo ("Table Missing [" + table_name + "]")
		return False

	def create_table_and_indexes (self, table_name, create_table_stmt, create_index_stmts):
		if self.connected_to_db:
			mycursor = self.mysql.cursor (buffered=True)
			mycursor.execute(create_table_stmt)
			self.sw.echo ("Created Table [" + table_name + "]")
			for index_stmt in create_index_stmts:
				mycursor.execute (index_stmt)
				self.sw.echo ("Created Index Using [" + index_stmt + "]")
				

	def get_max_value (self, tablename, colname, default_min):
		mycursor = self.mysql.cursor (buffered=True)
		mycursor.execute ('SELECT MAX(' + colname + ') FROM ' + tablename)
		myresult = mycursor.fetchall()
		for x in myresult:
			if x[0]:
				return x[0]
		return default_min

	def insert_rec (self, tablename, rec, pk_col_name):
		deletesql = "DELETE FROM " + tablename  + " WHERE Id = '" + rec[pk_col_name] + "'"
		mycursor = self.mysql.cursor (buffered=True)
		try:
			mycursor.execute (deletesql)
		except:
			self.sw.error ('Failed for: ' + deletesql)
			raise
		delCount = mycursor.rowcount

		insertsql = "INSERT INTO " + tablename + " (" + ",".join(rec.keys()) + ") VALUES ("
		arr = []
		for k in rec.keys():
			arr.append ('%s')
		insertsql += ",".join(arr) + ")"
		insertvals = []
		for colName in rec.keys():
			if isinstance(rec[colName], str):
				insertvals.append (str(rec[colName])[:255],)
			else:
				insertvals.append (rec[colName])
#		print (insertsql)
#		print (insertvals)
		try:
			mycursor.execute(insertsql, insertvals)
		except:
			self.sw.error ('Failed for: ' + insertsql)
			raise
		return delCount

	def commit (self):
		self.mysql.commit()



#from screenwriter import Screenwriter
from class_MySQL_Helper import MySQL_Helper
from class_SFDC_Helper  import SFDC_Helper
#from simple_salesforce  import Salesforce
import json


class Sync_Manager:

	def __init__ (self, p_sw, db_creds, sfdc_creds, control_map):
		self.sync_profile = {}
		self.sw = p_sw
		self.sf = SFDC_Helper (sfdc_creds, p_sw)
		self.mh = MySQL_Helper (db_creds, p_sw)
		default_min_dt = '2000-01-01' if 'min-SystemModstamp-date' not in control_map.keys() else control_map['min-SystemModstamp-date']
		self.min_default_ts = default_min_dt + 'T00:00:00.000+0000'
		self.dml_limit = 100 if 'dml-limit' not in control_map.keys() else control_map['dml-limit']
		self.fx_loaded = False

	def init_mysql (self):
		self.mh.create_database_if_not_exists ()

	def init_sfdc (self):
		self.sf.connect ()

	def add_sync_profile (self, table_name, columns, filters):
		prec = {}
		if 'Id' not in columns:
			columns.append ('Id')
		if 'SystemModstamp' not in columns:
			columns.append ('SystemModstamp')
		prec ['columns'] = columns
		prec ['filters'] = filters
		self.sync_profile [table_name] = prec

	def get_sql_datatype_from_sfdc_field (self, field):
		if field['name'].lower() == 'id':
			return 'CHAR (18) PRIMARY KEY'
		if field['type'] in ['id', 'reference']:
			return 'CHAR (18)'
		if field['type'] in ['boolean']:
			return 'CHAR (3)'
		if field['type'] in ['string', 'picklist', 'textarea', 'phone', 'email', 'url', 'multipicklist', 'encryptedstring']:
			if field['length'] <= 255:
				return 'CHAR (' + str(field['length']) + ')'
			else:
				return 'VARCHAR(' + str(field['length']) + ')'
		if field['type'] == 'address':
			return None
		if field['type'] in ['double', 'currency', 'location']:
			self.isdouble = True
			return 'DECIMAL (18, 3)'
		if field['type'] in ['datetime']:
			return 'CHAR (28)'
		if field['type'] in ['int', 'date']:
			return field['type'].upper()
		print ('ERROR! ' + field['name'] + ': INVALID TYPE: ' + str(field['type']) + ', ' + str(field['length']))
		exit(1)

	def create_missing_db_tables (self):
		if not self.mh.connected_to_db:
                        self.mh.connect_db ()
		for table_name in self.sync_profile.keys():
			if not self.mh.table_exists (table_name):
				prec = self.sync_profile[table_name]
				ddl_columns = []
				indexable_columns = []
				for field in self.sf.describe_object (table_name):
					if field['name'] in prec['columns']:
						ddl_columns.append (field['name'] + ' ' + self.get_sql_datatype_from_sfdc_field (field))
						if field['type'] in ['id', 'reference', 'datetime', 'date']:
							indexable_columns.append (field['name'])
				if 'USD' in prec.keys():
					ddl_columns.append (prec['USD'][2] + ' DECIMAL (18, 3)')
				ddl_stmt = 'CREATE TABLE ' + table_name + ' (\n' + ",\n".join(ddl_columns) + '\n)'
				index_stmts = []
				for i_col in indexable_columns:
					index_stmts.append ('CREATE INDEX idx_' + i_col + ' ON ' + table_name + ' (' + i_col + ')')
				self.mh.create_table_and_indexes (table_name, ddl_stmt, index_stmts)

	def add_USD_col (self, table_name, amt_col, curr_col, new_col):
		prec = self.sync_profile [table_name]
		prec ['USD'] = (amt_col, curr_col, new_col)
		self.sync_profile [table_name] = prec

	def sync_table (self, table_name):
		last_ts = str (self.mh.get_max_value (table_name, 'SystemModstamp', self.min_default_ts))
		self.sw.echo ("Syncing " + table_name + " [SystemModstamp >= " + last_ts + "]")
		prec = self.sync_profile [table_name]
		conditions = ['SystemModstamp >= ' + last_ts]
		for f in prec['filters']:
			conditions.append (f)
		self.sf.load_data (table_name, prec['columns'], conditions, [], self.dml_limit, 'SystemModstamp')
		if 'USD' in prec.keys():
			if not self.fx_loaded:
				self.sf.load_fx_rates ()
				self.fx_loaded = True
				self.sf.calculate_usd_column (table_name, prec['USD'][0], prec['USD'][1], prec['USD'][2])
		delete_count = 0
		unique_ts = set()
		for rec in self.sf.table_data [table_name]:
#			print (json.dumps (rec, default=lambda o: o.__dict__, sort_keys=True, indent=4))
			delete_count += self.mh.insert_rec (table_name, rec, 'Id')
			unique_ts.add (rec ['SystemModstamp'])
		num_synced = len(self.sf.table_data [table_name])
		self.sw.echo (table_name + " Inserted: " + str(num_synced - delete_count))
		self.sw.echo (table_name + " Updated:  " + str(delete_count))
		self.mh.commit ()
		if len(unique_ts) == 1 and num_synced == delete_count and num_synced == self.dml_limit:
			self.sw.error ("Repeat records selected.. This will turn into an infinite loop. Aborting!")
			exit (1)

	def sync (self):
		for table_name in self.sync_profile.keys():
			self.sync_table (table_name)
